@@ -10,6 +10,10 @@ from serial import SerialException
 
 from delumowave.config.delumo import (DEVICE, COMMAND, EEPROM_SIZE)
 
+# import threading
+from threading import Lock as threading_lock
+from multiprocessing import Lock as multiprocessing_lock
+
 _LOGGER = logging.getLogger(__name__)
 
 # Required delay between recording to UART
@@ -170,7 +174,7 @@ class DelumoWaveController:
                  parity: int = serial.PARITY_NONE,
                  stopbits: int = serial.STOPBITS_ONE,
                  timeout: float = 3.0,
-                 write_timeout: float = 3.0,
+                 # write_timeout: float = 3.0,
                  unlock_timeout: float = 5.0,
                  sleep_time: float = 0.1,
                  block: bool = False):
@@ -241,6 +245,10 @@ class DelumoWaveController:
         self.__block = block
         self.__node_version = None
         self.__last_uart = 0.0
+        self.__is_busy = False
+        # self.__lock = threading.Lock()
+        self.__threading_lock = threading_lock()
+        self.__multiprocessing_lock = multiprocessing_lock()
 
         self._controller = serial.serial_for_url(url=port_path,
                                                  baudrate=baudrate,
@@ -248,7 +256,7 @@ class DelumoWaveController:
                                                  parity=parity,
                                                  stopbits=stopbits,
                                                  timeout=timeout,
-                                                 write_timeout=write_timeout,
+                                                 # write_timeout=write_timeout,
                                                  # do_not_open=True,
                                                  exclusive=True)
 
@@ -488,47 +496,62 @@ class DelumoWaveController:
 
         # controller_is_blocked = True
         # time_to_wait = self.__unlock_timeout
-        # ==================== RADIO CONTROLLER START ====================
-        if not self._controller.is_open:
-            _LOGGER.info('Open serial port')
-            self._controller.open()
-        else:
-            _LOGGER.info('Serial port is already opened')
 
-        # No write to UART to wait for packet boundary
-        time_now = time()
-        sleep(UART_INTERVAL - (time_now - self.__last_uart) if time_now - self.__last_uart < UART_INTERVAL else 0.0)
+        with self.__threading_lock, self.__multiprocessing_lock:
+        # with self.__lock:
+            # ==================== RADIO CONTROLLER START ====================
+            if not self._controller.is_open:
+                _LOGGER.info('Open serial port')
+                self._controller.open()
+            else:
+                _LOGGER.info('Serial port is already opened')
 
-        # ----- COMMUNICATION START -----
-        # Send request
-        self._controller.reset_input_buffer()
-        self._controller.reset_output_buffer()
+            # No write to UART to wait for packet boundary
+            # time_now = time()
+            # sleep(UART_INTERVAL - (time_now - self.__last_uart) if time_now - self.__last_uart < UART_INTERVAL else 0.0)
 
-        try:
-            request_length = self._controller.write(request)
-        except SerialException as ex:
-            _LOGGER.exception(f'Error while sending request: {ex}')
-            raise
-        else:
-            # Check that all bytes were sent
-            if request_length != len(request):
-                raise RequestLengthError(request_length, len(request))
+            # Wait while radiocontroller is busy
+            # while self.__is_busy:
+            #     pass
+            # self.__is_busy = True
 
-        try:
-            request_length = self._controller.write(request)
-        except SerialException as ex:
-            _LOGGER.exception(f'Error while sending request: {ex}')
-            raise
-        else:
-            # Check that all bytes were sent
-            if request_length != len(request):
-                raise RequestLengthError(request_length, len(request))
-        finally:
-            self.__last_uart = time()
-        # ----- COMMUNICATION END -----
+            # ----- COMMUNICATION START -----
+            # Send request
+            self._controller.reset_input_buffer()
+            self._controller.reset_output_buffer()
 
-        if self.__block:
-            sleep(0.3)
+            # try:
+            #     request_length = self._controller.write(request)
+            #     self._controller.flush()
+            # except SerialException as ex:
+            #     _LOGGER.exception(f'Error while sending request: {ex}')
+            #     raise
+            # else:
+            #     # Check that all bytes were sent
+            #     if request_length != len(request):
+            #         raise RequestLengthError(request_length, len(request))
+
+            try:
+                request_length = self._controller.write(request)
+                self._controller.flush()
+            except SerialException as ex:
+                _LOGGER.exception(f'Error while sending request: {ex}')
+                raise
+            else:
+                # Check that all bytes were sent
+                if request_length != len(request):
+                    raise RequestLengthError(request_length, len(request))
+            finally:
+                self.__last_uart = time()
+            # ----- COMMUNICATION END -----
+
+            if self.__block:
+                sleep(0.3)
+
+            sleep(0.08)
+            while self._controller.out_waiting != 0:
+                pass
+            # self.__is_busy = False
 
         return True
 
@@ -570,49 +593,57 @@ class DelumoWaveController:
 
         _LOGGER.debug(f'Request: {hex_repr(request)}')
 
-        # ==================== RADIO CONTROLLER START ====================
-        if not self._controller.is_open:
-            _LOGGER.info('Open serial port')
-            self._controller.open()
-        else:
-            _LOGGER.info('Serial port is already opened')
+        with self.__threading_lock, self.__multiprocessing_lock:
+        # with self.__lock:
+            # ==================== RADIO CONTROLLER START ====================
+            if not self._controller.is_open:
+                _LOGGER.info('Open serial port')
+                self._controller.open()
+            else:
+                _LOGGER.info('Serial port is already opened')
 
-        # No write to UART to wait for packet boundary
-        time_now = time()
-        sleep(UART_INTERVAL - (time_now - self.__last_uart) if time_now - self.__last_uart < UART_INTERVAL else 0.0)
+            # No write to UART to wait for packet boundary
+            # time_now = time()
+            # sleep(UART_INTERVAL - (time_now - self.__last_uart) if time_now - self.__last_uart < UART_INTERVAL else 0.0)
 
-        # ----- COMMUNICATION START -----
-        # Send request
-        self._controller.reset_input_buffer()
-        self._controller.reset_output_buffer()
+            # ----- COMMUNICATION START -----
+            # Send request
+            self._controller.reset_input_buffer()
+            self._controller.reset_output_buffer()
 
-        try:
-            request_length = self._controller.write(request)
-        except SerialException as ex:
-            _LOGGER.exception(f'Error while sending request: {ex}')
-            raise
-        else:
-            # Check that all bytes were sent
-            if request_length != len(request):
-                raise RequestLengthError(request_length, len(request))
+            try:
+                request_length = self._controller.write(request)
+                self._controller.flush()
+            except SerialException as ex:
+                _LOGGER.exception(f'Error while sending request: {ex}')
+                raise
+            else:
+                # Check that all bytes were sent
+                if request_length != len(request):
+                    raise RequestLengthError(request_length, len(request))
 
-        # Receive response
-        try:
-            time1 = time()
-            response = self._controller.read(expected_response_length)
-            time2 = time()
-        except SerialException as ex:
-            #     _LOGGER.exception(f'Unknown error in serial port while '
-            #                       f'receiving response: {ex}')
-            raise
-        else:
-            # Check that all bytes were received correctly
-            self._check_response(request, response, expected_response_length)
-        # ----- COMMUNICATION END -----
+            # Receive response
+            try:
+                time1 = time()
+                response = self._controller.read(expected_response_length)
+                self._controller.flush()
+                time2 = time()
+            except SerialException as ex:
+                #     _LOGGER.exception(f'Unknown error in serial port while '
+                #                       f'receiving response: {ex}')
+                raise
+            else:
+                # Check that all bytes were received correctly
+                self._check_response(request, response, expected_response_length)
+            # ----- COMMUNICATION END -----
 
-        if self.__block:
-            sleep(0.1)
-        # ==================== RADIO CONTROLLER END ======================
+            if self.__block:
+                sleep(0.1)
+
+            # sleep(0.1)
+            while self._controller.out_waiting != 0:
+                pass
+            # ==================== RADIO CONTROLLER END ======================
 
         _LOGGER.debug(f'Response transfer time: {time2 - time1} sec')
         _LOGGER.debug(f'Request length -> {request_length}. '
@@ -660,48 +691,56 @@ class DelumoWaveController:
 
         _LOGGER.debug(f'Request: {hex_repr(request)}')
 
-        # ==================== RADIO CONTROLLER START ====================
-        if not self._controller.is_open:
-            _LOGGER.info('Open serial port')
-            self._controller.open()
-        else:
-            _LOGGER.info('Serial port is already opened')
+        with self.__threading_lock, self.__multiprocessing_lock:
+        # with self.__lock:
+            # ==================== RADIO CONTROLLER START ====================
+            if not self._controller.is_open:
+                _LOGGER.info('Open serial port')
+                self._controller.open()
+            else:
+                _LOGGER.info('Serial port is already opened')
 
-        # No write to UART to wait for packet boundary
-        time_now = time()
-        sleep(UART_INTERVAL - (time_now - self.__last_uart) if time_now - self.__last_uart < UART_INTERVAL else 0.0)
+            # No write to UART to wait for packet boundary
+            time_now = time()
+            sleep(UART_INTERVAL - (time_now - self.__last_uart) if time_now - self.__last_uart < UART_INTERVAL else 0.0)
 
-        # ----- COMMUNICATION START -----
-        # Send request
-        self._controller.reset_input_buffer()
-        self._controller.reset_output_buffer()
+            # ----- COMMUNICATION START -----
+            # Send request
+            self._controller.reset_input_buffer()
+            self._controller.reset_output_buffer()
 
-        try:
-            request_length = self._controller.write(request)
-        except SerialException as ex:
-            _LOGGER.exception(f'Error while sending request: {ex}')
-            raise
-        else:
-            # Check that all bytes were sent
-            if request_length != len(request):
-                raise RequestLengthError(request_length, len(request))
+            # try:
+            #     request_length = self._controller.write(request)
+            #     self._controller.flush()
+            # except SerialException as ex:
+            #     _LOGGER.exception(f'Error while sending request: {ex}')
+            #     raise
+            # else:
+            #     # Check that all bytes were sent
+            #     if request_length != len(request):
+            #         raise RequestLengthError(request_length, len(request))
 
-        try:
-            request_length = self._controller.write(request)
-        except SerialException as ex:
-            _LOGGER.exception(f'Error while sending request: {ex}')
-            raise
-        else:
-            # Check that all bytes were sent
-            if request_length != len(request):
-                raise RequestLengthError(request_length, len(request))
-        finally:
-            self.__last_uart = time()
-        # ----- COMMUNICATION END -----
+            try:
+                request_length = self._controller.write(request)
+                self._controller.flush()
+            except SerialException as ex:
+                _LOGGER.exception(f'Error while sending request: {ex}')
+                raise
+            else:
+                # Check that all bytes were sent
+                if request_length != len(request):
+                    raise RequestLengthError(request_length, len(request))
+            finally:
+                self.__last_uart = time()
+            # ----- COMMUNICATION END -----
 
-        if self.__block:
-            expected_block_time = 0.3 + request_length / 280
-            sleep(expected_block_time)
-        # ==================== RADIO CONTROLLER END ======================
+            if self.__block:
+                expected_block_time = 0.3 + request_length / 280
+                sleep(expected_block_time)
+
+            sleep(0.08)
+            while self._controller.out_waiting != 0:
+                pass
+            # ==================== RADIO CONTROLLER END ======================
 
         return True
